import datetime
import typing
import json

from gumo.core.injector import injector
from gumo.core import EntityKey
from gumo.core import NoneKey
from gumo.datastore.infrastructure.entity_key_mapper import EntityKeyMapper

from google.cloud import datastore


class DatastoreMapperMixin:
    _entity_key_mapper = None
    DatastoreEntity = datastore.Entity

    @property
    def entity_key_mapper(self) -> EntityKeyMapper:
        if self._entity_key_mapper is None:
            self._entity_key_mapper = injector.get(EntityKeyMapper)  # type: EntityKeyMapper

        return self._entity_key_mapper

    def to_entity_key(self, datastore_key: typing.Optional[datastore.Key]) -> EntityKey:
        return self.entity_key_mapper.to_entity_key(datastore_key=datastore_key)

    def to_datastore_key(self, entity_key: typing.Union[EntityKey, NoneKey, None]) -> typing.Optional[datastore.Key]:
        return self.entity_key_mapper.to_datastore_key(entity_key=entity_key)

    def convert_datetime(self, t: datetime.datetime) -> typing.Optional[datetime.datetime]:
        if t is None:
            return

        return datetime.datetime(
            year=t.year,
            month=t.month,
            day=t.day,
            hour=t.hour,
            minute=t.minute,
            second=t.second,
            microsecond=t.microsecond,
            tzinfo=datetime.timezone.utc,
        )

    def loads_json_property(self, b: bytes) -> typing.Union[typing.List, typing.Dict]:
        return json.loads(b.decode('utf-8'))

    def dumps_json_property(self, doc: typing.Dict) -> bytes:
        return json.dumps(doc).encode('utf-8')

    def to_datastore_entity(self, entity: typing.Any) -> datastore.Entity:
        raise NotImplementedError()

    def to_domain_entity(self, doc: datastore.Entity) -> typing.Any:
        raise NotImplementedError()



import dataclasses
import datetime
import typing

from google.cloud import datastore


@dataclasses.dataclass()
class DataModel:
    key: datastore.Key

    exclude_from_indexes: typing.ClassVar[typing.List[str]] = []
    DatastoreEntity: typing.ClassVar = datastore.Entity
    DatastoreKey: typing.ClassVar = datastore.Key

    def to_datastore_entity(self) -> "DataModel.DatastoreEntity":
        raise NotImplementedError()

    @classmethod
    def from_datastore_entity(cls, doc: DatastoreEntity) -> "DataModel":
        raise NotImplementedError()

    @classmethod
    def convert_optional_datetime(
        cls, t: typing.Optional[datetime.datetime]
    ) -> typing.Optional[datetime.datetime]:
        if t is None:
            return None

        return cls.convert_datetime(t)

    @classmethod
    def convert_datetime(cls, t: datetime.datetime) -> datetime.datetime:
        return datetime.datetime(
            year=t.year,
            month=t.month,
            day=t.day,
            hour=t.hour,
            minute=t.minute,
            second=t.second,
            microsecond=t.microsecond,
            tzinfo=datetime.timezone.utc,
        )



from gumo.core.application.entity_key import KeyIDAllocator
from gumo.datastore.infrastructure.key_id_allocator import DatastoreKeyIDAllocator


def datastore_binder(binder):
    binder.bind(KeyIDAllocator, to=DatastoreKeyIDAllocator)



from injector import inject

from typing import Optional
from typing import Union
from google.cloud import datastore

from gumo.core import GumoConfiguration

from gumo.core import EntityKey
from gumo.core import NoneKey
from gumo.core import EntityKeyFactory


class EntityKeyMapper:
    @inject
    def __init__(
            self,
            gumo_config: GumoConfiguration,
            entity_key_factory: EntityKeyFactory,
    ):
        self._gumo_config = gumo_config
        self._entity_key_factory = entity_key_factory

    def to_entity_key(self, datastore_key: Optional[datastore.Key]) -> EntityKey:
        if datastore_key is None:
            return NoneKey.get_instance()

        entity_key = self._entity_key_factory.build_from_pairs(pairs=datastore_key.path)
        return entity_key

    def to_datastore_key(self, entity_key: Union[EntityKey, NoneKey, None]) -> Optional[datastore.Key]:
        if entity_key is None or isinstance(entity_key, NoneKey):
            return None

        project = self._gumo_config.google_cloud_project.value
        datastore_key = datastore.Key(*entity_key.flat_pairs(), project=project)

        return datastore_key



import typing
from contextlib import contextmanager

from gumo.core.injector import injector
from gumo.core import EntityKey
from gumo.core import NoneKey
from gumo.datastore.infrastructure.configuration import DatastoreConfiguration
from gumo.datastore.infrastructure.entity_key_mapper import EntityKeyMapper

from google.cloud import datastore


class DatastoreRepositoryMixin:
    _datastore_client = None
    _entity_key_mapper = None

    DatastoreEntity = datastore.Entity

    @property
    def datastore_client(self) -> datastore.Client:
        if self._datastore_client is None:
            configuration = injector.get(DatastoreConfiguration)  # type: DatastoreConfiguration
            self._datastore_client = configuration.client

        return self._datastore_client

    @property
    def entity_key_mapper(self) -> EntityKeyMapper:
        if self._entity_key_mapper is None:
            self._entity_key_mapper = injector.get(EntityKeyMapper)  # type: EntityKeyMapper

        return self._entity_key_mapper

    def to_entity_key(self, datastore_key: typing.Optional[datastore.Key]) -> EntityKey:
        return self.entity_key_mapper.to_entity_key(datastore_key=datastore_key)

    def to_datastore_key(self, entity_key: typing.Union[EntityKey, NoneKey, None]) -> typing.Optional[datastore.Key]:
        return self.entity_key_mapper.to_datastore_key(entity_key=entity_key)


@contextmanager
def datastore_transaction():
    datastore_client = injector.get(DatastoreConfiguration).client  # type: datastore.Client

    with datastore_client.transaction():
        yield
